import os

from aws_infrastructure_sdk.cloud_formation.custom_resources import custom_root_dir
from aws_infrastructure_sdk.cloud_formation.custom_resources.service.abstract_custom_service import AbstractCustomService
from troposphere.iam import Role, Policy


class CognitoDomainService(AbstractCustomService):
    def __init__(self):
        super().__init__()

        self.src = os.path.join(
            custom_root_dir,
            'package',
            'cognito_domain'
        )

        self.lambda_handler = 'index.handler'
        self.lambda_runtime = 'python3.6'
        self.lambda_memory = 128
        self.lambda_timeout = 60
        self.lambda_name = 'CfCustomResourceCognitoUserPoolDomainName'
        self.lambda_description = (
            'Lambda function enabling AWS CloudFormation '
            'to create domain name for AWS Cognito user pool.'
        )

    def role(self) -> Role:
        return Role(
            "CfCustomResourceCognitoUserPoolDomainNameLambdaRole",
            Path="/",
            Policies=[Policy(
                PolicyName="CfCustomResourceCognitoUserPoolDomainNameLambdaPolicy",
                PolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [{
                        "Action": ["logs:*"],
                        "Resource": "arn:aws:logs:*:*:*",
                        "Effect": "Allow"
                    }, {
                        "Action": [
                            "cognito-idp:CreateUserPoolDomain",
                            "cognito-idp:UpdateUserPoolDomain",
                            "cognito-idp:DeleteUserPoolDomain",
                        ],
                        "Resource": "*",
                        "Effect": "Allow"
                    }]
                })],
            AssumeRolePolicyDocument={"Version": "2012-10-17", "Statement": [
                {
                    "Action": ["sts:AssumeRole"],
                    "Effect": "Allow",
                    "Principal": {
                        "Service": [
                            "lambda.amazonaws.com",
                        ]
                    }
                }
            ]},
        )



from troposphere import AWSObject
from troposphere.validators import boolean, positive_integer


class CustomUserPoolClient(AWSObject):
    resource_type = "AWS::Cognito::UserPoolClient"

    props = {
        'ClientName': (str, False),
        'ExplicitAuthFlows': ([str], False),
        'GenerateSecret': (boolean, False),
        'ReadAttributes': ([str], False),
        'RefreshTokenValidity': (positive_integer, False),
        'UserPoolId': (str, True),
        'WriteAttributes': ([str], False),
        'CallbackURLs': ([str], False),
        'DefaultRedirectURI': (str, False),
        'LogoutURLs': ([str], False),
        'SupportedIdentityProviders': ([str], False),
        'AllowedOAuthScopes': ([str], False),
        'AllowedOAuthFlows': ([str], False),
        'AllowedOAuthFlowsUserPoolClient': (bool, False),
    }



from typing import List, Dict
from troposphere import Template
from troposphere.ec2 import SecurityGroup, Subnet, VPC
from troposphere.s3 import Bucket

from aws_infrastructure_sdk.cloud_formation.fargate_ci_cd.ecs_autoscaling import Autoscaling
from aws_infrastructure_sdk.cloud_formation.fargate_ci_cd.ecs_loadbalancer import Loadbalancing
from aws_infrastructure_sdk.cloud_formation.fargate_ci_cd.ecs_main import Ecs
from aws_infrastructure_sdk.cloud_formation.fargate_ci_cd.ecs_pipeline import EcsPipeline
from aws_infrastructure_sdk.cloud_formation.types import AwsRef


class EcsParams:
    """
    Parameters class which specifies deployed container and ecs parameters such as name, port, etc.
    """
    def __init__(
            self,
            container_name: str,
            container_cpu: str,
            container_ram: str,
            container_port: int,
            container_environment: Dict[str, AwsRef],
            ecs_security_groups: List[SecurityGroup],
            ecs_subnets: List[Subnet]
    ) -> None:
        """
        Constructor.

        :param container_name: The name that will be given to a newly deployed container.
        :param container_cpu: Cpu points for the deployed container. 1 CPU = 1024 Cpu units.
        :param container_ram: Memory for the deployed container. 1 GB Ram = 1024 units.
        :param container_port: An open container port through which a loadbalancer can communicate.
        :param container_environment: Environment that will be passed to a running container.
        :param ecs_security_groups: Security groups for ecs service in which containers are placed.
        :param ecs_subnets: Subnets to which new containers will be deployed.
        """
        self.container_name = container_name
        self.container_cpu = container_cpu
        self.container_ram = container_ram
        self.container_port = container_port
        self.container_environment = container_environment
        self.ecs_security_groups = ecs_security_groups
        self.ecs_subnets = ecs_subnets


class LoadBalancerParams():
    def __init__(self, subnets: List[Subnet], security_groups: List[SecurityGroup], dns: str):
        """
        Constructor.

        :param subnet: Subnets in which a newly created loadbalancer can operate.
        :param dns: A domain name for a loadbalancer. E.g. myweb.com. This is used to issue a new
        certificate in order a loadbalancer can use HTTPS connections.
        """
        self.dns = dns
        self.security_groups = security_groups
        self.lb_subnets = subnets


class PipelineParams:
    """
    Parameters class which specifies various parameters for ci/cd pipeline.
    """
    def __init__(self, artifact_builds_bucket: Bucket):
        """
        Constructor.

        :param artifact_builds_bucket: An artifacts bucket which will be used by a ci/cd pipeline to write
        and read build/source artifacts.
        """
        self.artifact_builds_bucket = artifact_builds_bucket


class Main:
    """
    Main class which provisions whole infrastructure regarding:
        1. ECS Fargate,
        2. Autoscaling,
        3. Loadbalancing,
        4. Pipeline CI/CD.

    Known pitfalls:
        1. Cloudformation updates a deployment group or an ecs service by deleting it and creating a new one. However
        "create" action is never called (probably a bug in cloudformation dealing with custom resources functionality).
        After an update a deployment group or an ecs service are deleted an no longer exist. To fix that - tear down
        the whole main ecs fargate ci/cd stack and rerun it.
    """
    def __init__(
            self,
            prefix: str,
            region: str,
            account_id: str,
            vpc: VPC,
            lb_params: LoadBalancerParams,
            ecs_params: EcsParams,
            pipeline_params: PipelineParams,
    ) -> None:
        """
        Constructor.

        :param prefix: The prefix for all newly created resources. E.g. Wordpress.
        :param region: The region where resources and the stack are deployed.
        :param account_id: The id of the account which executes this stack.
        :param vpc: Virtual private cloud (VPC).
        :param lb_params: Loadbalancer parameters.
        :param ecs_params: Compute power parameters for newly deployed container.
        :param pipeline_params: Parameters for a ci/cd pipeline.
        """
        # Create a loadbalancer to which an ecs service will attach.
        self.load_balancing = Loadbalancing(
            prefix=prefix,
            subnets=lb_params.lb_subnets,
            lb_security_groups=lb_params.security_groups,
            vpc=vpc,
            desired_domain_name=lb_params.dns
        )

        # Create main fargate ecs service.
        self.ecs = Ecs(
            prefix=prefix,
            aws_region=region,
            environment=ecs_params.container_environment,
            cpu=ecs_params.container_cpu,
            ram=ecs_params.container_ram,
            container_name=ecs_params.container_name,
            container_port=ecs_params.container_port,
            target_group=self.load_balancing.target_group_1_http,
            security_groups=ecs_params.ecs_security_groups,
            subnets=ecs_params.ecs_subnets,
            # Ecs can not be created until a loadbalancer is created.
            depends_on_loadbalancers=[self.load_balancing.load_balancer],
            # Ecs can not be created until target groups are created.
            depends_on_target_groups=[
                self.load_balancing.target_group_1_http,
                self.load_balancing.target_group_2_http
            ],
            # Ecs can not be created until listeners are associated with a loadbalancer and target groups.
            depends_on_listeners=[
                self.load_balancing.listener_http_1,
                self.load_balancing.listener_http_2,
                self.load_balancing.listener_https_1,
                self.load_balancing.listener_https_2,
            ]
        )

        # Create autoscaling for an ecs service.
        self.autoscaling = Autoscaling(
            prefix=prefix,
            service_name=self.ecs.service.ServiceName,
            cluster_name=self.ecs.cluster.ClusterName,
            service_resource_name=self.ecs.service.title
        )

        # Finally, create a pipeline deploying from ECR to ECS with Blue/Green deployments style.
        self.pipeline = EcsPipeline(
            prefix=prefix,
            main_target_group=self.load_balancing.target_group_1_http,
            deployments_target_group=self.load_balancing.target_group_2_http,
            main_listener=self.load_balancing.listener_https_1,
            deployments_listener=self.load_balancing.listener_https_2,
            ecs_service=self.ecs.service,
            ecs_cluster=self.ecs.cluster,
            artifact_builds_s3=pipeline_params.artifact_builds_bucket,
            task_def=self.ecs.create_task_def(),
            app_spec=self.ecs.create_appspec(),
            aws_account_id=account_id,
            aws_region=region,
            # Pipeline can not be created until the ecs service itself is created.
        )

    def add(self, template: Template):
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        self.load_balancing.add(template)
        self.ecs.add(template)
        self.autoscaling.add(template)
        self.pipeline.add(template)



import boto3

from typing import Any, Dict

client = boto3.client('codedeploy')


class Group:
    @staticmethod
    def create(**kwargs) -> Dict[str, Any]:
        kwargs = dict(
            applicationName=kwargs.get('applicationName'),
            deploymentGroupName=kwargs.get('deploymentGroupName'),
            deploymentConfigName=kwargs.get('deploymentConfigName'),
            ec2TagFilters=kwargs.get('ec2TagFilters'),
            onPremisesInstanceTagFilters=kwargs.get('onPremisesInstanceTagFilters'),
            autoScalingGroups=kwargs.get('autoScalingGroups'),
            serviceRoleArn=kwargs.get('serviceRoleArn'),
            triggerConfigurations=kwargs.get('triggerConfigurations'),
            alarmConfiguration=kwargs.get('alarmConfiguration'),
            autoRollbackConfiguration=kwargs.get('autoRollbackConfiguration'),
            deploymentStyle=kwargs.get('deploymentStyle'),
            blueGreenDeploymentConfiguration=kwargs.get('blueGreenDeploymentConfiguration'),
            loadBalancerInfo=kwargs.get('loadBalancerInfo'),
            ec2TagSet=kwargs.get('ec2TagSet'),
            ecsServices=kwargs.get('ecsServices'),
            onPremisesTagSet=kwargs.get('onPremisesTagSet'),
            tags=kwargs.get('tags')
        )

        kwargs = {key: value for key, value in kwargs.items() if key and value}
        return client.create_deployment_group(**kwargs)

    @staticmethod
    def update(**kwargs: Dict[str, Any]) -> Dict[str, Any]:
        kwargs = dict(
            applicationName=kwargs.get('applicationName'),
            currentDeploymentGroupName=kwargs.get('deploymentGroupName'),
            newDeploymentGroupName=kwargs.get('newDeploymentGroupName'),
            deploymentConfigName=kwargs.get('deploymentConfigName'),
            ec2TagFilters=kwargs.get('ec2TagFilters'),
            onPremisesInstanceTagFilters=kwargs.get('onPremisesInstanceTagFilters'),
            autoScalingGroups=kwargs.get('autoScalingGroups'),
            serviceRoleArn=kwargs.get('serviceRoleArn'),
            triggerConfigurations=kwargs.get('triggerConfigurations'),
            alarmConfiguration=kwargs.get('alarmConfiguration'),
            autoRollbackConfiguration=kwargs.get('autoRollbackConfiguration'),
            deploymentStyle=kwargs.get('deploymentStyle'),
            blueGreenDeploymentConfiguration=kwargs.get('blueGreenDeploymentConfiguration'),
            loadBalancerInfo=kwargs.get('loadBalancerInfo'),
            ec2TagSet=kwargs.get('ec2TagSet'),
            ecsServices=kwargs.get('ecsServices'),
            onPremisesTagSet=kwargs.get('onPremisesTagSet'),
        )

        kwargs = {key: value for key, value in kwargs.items() if key and value}
        return client.update_deployment_group(**kwargs)

    @staticmethod
    def delete(**kwargs: Dict[str, Any]) -> Dict[str, Any]:
        kwargs = dict(
            applicationName=kwargs.get('applicationName'),
            deploymentGroupName=kwargs.get('deploymentGroupName'),
        )

        kwargs = {key: value for key, value in kwargs.items() if key and value}
        return client.delete_deployment_group(**kwargs)



from troposphere import AWSObject


class CustomS3Notification(AWSObject):
    resource_type = "Custom::S3Notification"

    props = {
        'ServiceToken': (str, True),
        'BucketName': (str, True),
        'NotificationConfiguration': (dict, True),
    }



from troposphere import Template, GetAtt, Ref
from troposphere.awslambda import Code, Function, Environment
from troposphere.iam import Role, Policy
from aws_infrastructure_sdk.cloud_formation.lambda_ci_cd import lambda_ci_cd_root


class DeploymentLambda:
    """
    A class which provisions an infrastructure for an AWS Lambda function which executes another AWS Lambda
    function deployment from a S3 bucket in which a deployment package resides.
    """
    def __init__(self, prefix: str, lambda_under_deployment: Function) -> None:
        """
        Constructor.

        :param prefix: A prefix for deployment lambda resource names.
        :param lambda_under_deployment: An AWS Lambda function to execute deployments against.
        """
        self.lambda_role = Role(
            prefix + "DeploymentLambdaRole",
            Path="/",
            Policies=[Policy(
                PolicyName=prefix + "DeploymentLambdaRole",
                PolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [{
                        "Action": ["logs:*"],
                        "Resource": "arn:aws:logs:*:*:*",
                        "Effect": "Allow"
                    }, {
                        "Action": ["lambda:UpdateFunctionCode"],
                        "Resource": "*",
                        "Effect": "Allow"
                    }, {
                        "Action": ["s3:*"],
                        "Resource": "*",
                        "Effect": "Allow"
                    }]
                })],
            AssumeRolePolicyDocument={"Version": "2012-10-17", "Statement": [
                {
                    "Action": ["sts:AssumeRole"],
                    "Effect": "Allow",
                    "Principal": {
                        "Service": [
                            "lambda.amazonaws.com",
                        ]
                    }
                }
            ]},
        )

        self.function = Function(
            prefix + "DeploymentLambda",
            Code=Code(ZipFile=self.__read_template()),
            Handler='index.handler',
            Role=GetAtt(self.lambda_role, "Arn"),
            Runtime='python3.6',
            MemorySize='128',
            FunctionName=prefix + 'DeploymentLambda',
            Timeout='10',
            Environment=Environment(
                Variables={
                    'LAMBDA_FUNCTION_NAME': Ref(lambda_under_deployment)
                }
            ),
            Description=(
                f'Deployment lambda which updates lambda under deployment function code '
                f'from an output from a ci/cd pipeline for {prefix.lower()}.'
            )
        )

    def add(self, template: Template) -> None:
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        template.add_resource(self.function)
        template.add_resource(self.lambda_role)

    @staticmethod
    def __read_template() -> str:
        """
        Reads and returns function code for a deployment AWS Lambda function.

        :return: Source code as a string.
        """
        with open(f'{lambda_ci_cd_root}/deployment.template', 'r') as file:
            return ''.join(file.readlines())



from abc import ABC, abstractmethod
from aws_infrastructure_sdk.cloud_formation.types import AwsRef
from troposphere import GetAtt, Template
from troposphere.awslambda import Function, Code
from troposphere.iam import Role
from aws_infrastructure_sdk.cloud_formation.custom_resources.config import CF_CUSTOM_BUILDS_BUCKET


class AbstractCustomService(ABC):
    def __init__(self):
        self.src = None
        self.lambda_handler = None
        self.lambda_runtime = None
        self.lambda_memory = None
        self.lambda_timeout = None
        self.lambda_description = None
        self.lambda_name = None

    @abstractmethod
    def role(self) -> Role:
        pass

    def service_token(self) -> AwsRef:
        return GetAtt(self.function(), 'Arn')

    def function(self) -> Function:
        assert self.lambda_name
        assert self.lambda_handler
        assert self.lambda_runtime
        assert self.lambda_memory
        assert self.lambda_timeout
        assert self.lambda_description

        return Function(
            self.lambda_name,
            Code=Code(
                S3Bucket=CF_CUSTOM_BUILDS_BUCKET,
                S3Key=self.lambda_name
            ),
            Handler=self.lambda_handler,
            Role=GetAtt(self.role(), "Arn"),
            Runtime=self.lambda_runtime,
            MemorySize=self.lambda_memory,
            FunctionName=self.lambda_name,
            Timeout=self.lambda_timeout,
            Description=self.lambda_description
        )

    def add(self, template: Template):
        template.add_resource(self.function())
        template.add_resource(self.role())



import copy

from typing import Any, Dict


class Fixer:
    def __init__(self, kwargs: Dict[str, Any]):
        self.__kwargs = kwargs

    def get(self) -> Dict[str, Any]:
        return copy.deepcopy(self.__kwargs)

    def fix(self) -> 'Fixer':
        """
        Fixes lambda event resource parameters to comply with boto3 library parameters.

        :return: No return.
        """
        # All of the resource properties come with capitalized keys. Boto3 parameters start from lowercase.
        # Make sure to convert uppercase first char to a lowercase.
        self.__kwargs = {self.__inverse_capitalize(key): value for key, value in self.__kwargs.items()}

        # File contents must be bytes.
        if self.__kwargs.get('putFiles'):
            for put_file in self.__kwargs['putFiles']:
                if put_file.get('fileContent'):
                    file_content = put_file['fileContent']
                    if isinstance(file_content, str):
                        put_file['fileContent'] = file_content.encode()

        return self

    @staticmethod
    def __inverse_capitalize(string: str) -> str:
        string = list(string)
        first = string[0].lower()
        string[0] = first
        return ''.join(string)



from troposphere.cloudformation import AWSCustomObject


class CustomCognitoClientSecret(AWSCustomObject):
    resource_type = "Custom::CognitoClientSecret"

    props = {
        'ServiceToken': (str, True),
        'UserPoolId': (str, True),
        'ClientId': (str, True),
    }



from typing import List, Dict, Any
from troposphere import Template
from troposphere.ec2 import SecurityGroup, Subnet
from troposphere.iam import Role
from troposphere.s3 import Bucket
from aws_infrastructure_sdk.cloud_formation.lambda_ci_cd.lambda_function import LambdaFunction
from aws_infrastructure_sdk.cloud_formation.lambda_ci_cd.lambda_pipeline import LambdaPipeline


class LambdaParams:
    """
    AWS Lambda Function parameters that define compute power and other parameters.
    """
    def __init__(
            self,
            description: str,
            memory: int,
            timeout: int,
            handler: str,
            runtime: str,
            role: Role,
            env: Dict[Any, Any],
            security_groups: List[SecurityGroup],
            subnets: List[Subnet]
    ) -> None:
        """
        Constructor.

        :param description: Description of the lambda function.
        :param memory: Memory units for the lambda function.
        :param timeout: Time after which lambda function is terminated if it does not complete.
        :param handler: The name of the main function which will be called to interact with lambda function.
        :param runtime: The type of lambda runtime e.g. python3.6.
        :param role: Role resource that defines what lambda can and can not do.
        :param env: OS-level environment variables for the function.
        :param security_groups: Security groups for the function.
        :param subnets: Subnets in which the function lives. Note, subnets must have NAT Gateway.
        """
        self.role = role
        self.runtime = runtime
        self.handler = handler
        self.timeout = timeout
        self.memory = memory
        self.description = description
        self.env = env
        self.security_groups = security_groups
        self.subnets = subnets


class PipelineParams:
    """
    Parameters class which specifies various parameters for ci/cd pipeline.
    """
    def __init__(self, artifact_builds_bucket: Bucket):
        """
        Constructor.

        :param artifact_builds_bucket: An artifacts bucket which will be used by a ci/cd pipeline to write
        and read build/source artifacts.
        """
        self.artifact_builds_bucket = artifact_builds_bucket


class Main:
    """
    A main project to create a serverless AWS Lambda function with complete CI/CD integration.

    Known pitfalls:
        1. Currently supported only for python.
        2. If your lambda needs internet access - subnets must have a NAT Gateway attached.
    """
    def __init__(
            self,
            prefix: str,
            lambda_params: LambdaParams,
            pipeline_params: PipelineParams
    ):
        self.function = LambdaFunction(
            prefix=prefix,
            description=lambda_params.description,
            memory=lambda_params.memory,
            timeout=lambda_params.timeout,
            handler=lambda_params.handler,
            runtime=lambda_params.runtime,
            role=lambda_params.role,
            env=lambda_params.env,
            security_groups=lambda_params.security_groups,
            subnets=lambda_params.subnets
        )

        self.pipeline = LambdaPipeline(
            prefix=prefix,
            lambda_under_deployment=self.function.lambda_function,
            artifacts_bucket=pipeline_params.artifact_builds_bucket
        )

    def add(self, template: Template) -> None:
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        self.function.add(template)
        self.pipeline.add(template)



from troposphere import GetAtt, Template
from troposphere.awslambda import Function, Permission
from troposphere.codebuild import Project, Artifacts, Environment, Source
from troposphere.codecommit import Repository
from troposphere.codepipeline import *
from troposphere.iam import Role, Policy
from troposphere.s3 import Bucket, PublicAccessBlockConfiguration
from aws_infrastructure_sdk.cloud_formation.custom_resources.resource.s3_notification import CustomS3Notification
from aws_infrastructure_sdk.cloud_formation.custom_resources.service.s3_notification import S3NotificationService
from aws_infrastructure_sdk.cloud_formation.lambda_ci_cd import lambda_ci_cd_root
from aws_infrastructure_sdk.cloud_formation.lambda_ci_cd.deployment_lambda import DeploymentLambda


class LambdaPipeline:
    """
    A main ci/cd pipeline class.
    """
    RUNTIME_TO_BUILD_SPECS_MAP = {
        'python3.6': 'python-buildspec.yml',
        'python3.7': 'python-buildspec.yml',
    }

    def __init__(
            self,
            prefix: str,
            lambda_under_deployment: Function,
            artifacts_bucket: Bucket
    ) -> None:
        """
        Constructor.

        :param prefix: A prefix for pipeline resource names.
        :param lambda_under_deployment: Main AWS Lambda function for which the pipeline should be created.
        :param artifacts_bucket: An S3 bucket in which pipeline builds are saved and read.
        """
        self.lambda_under_deployment = lambda_under_deployment

        self.deployment_group_role = Role(
            prefix + 'DeploymentGroupRole',
            Path='/',
            Policies=[Policy(
                PolicyName=prefix + 'FargateEcsDeploymentGroupPolicy',
                PolicyDocument={
                    'Version': '2012-10-17',
                    'Statement': [{
                        'Action': [
                            "lambda:InvokeFunction",
                            "cloudwatch:DescribeAlarms",
                            "sns:Publish",
                            "s3:GetObject",
                            "s3:GetObjectMetadata",
                            "s3:GetObjectVersion",
                            "iam:PassRole"
                        ],
                        "Resource": "*",
                        "Effect": "Allow"
                    }]
                })],
            AssumeRolePolicyDocument={'Version': '2012-10-17', 'Statement': [
                {
                    'Action': ['sts:AssumeRole'],
                    'Effect': 'Allow',
                    'Principal': {
                        'Service': [
                            'codedeploy.amazonaws.com',
                        ]
                    }
                }
            ]},
        )

        self.code_build_role = Role(
            prefix + 'CodeBuildRole',
            Path='/',
            Policies=[Policy(
                PolicyName=prefix + 'CodeBuildPolicy',
                PolicyDocument={
                    'Version': '2012-10-17',
                    'Statement': [{
                        'Effect': 'Allow',
                        'Action': ['codebuild:*'],
                        'Resource': '*'
                    }, {
                        "Effect": "Allow",
                        "Action": ["logs:*"],
                        "Resource": "arn:aws:logs:*:*:*"
                    }, {
                        'Effect': 'Allow',
                        'Action': ['s3:*'],
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': ['ssm:*'],
                        'Resource': '*'
                    }]
                })],
            AssumeRolePolicyDocument={'Version': '2012-10-17', 'Statement': [
                {
                    'Action': ['sts:AssumeRole'],
                    'Effect': 'Allow',
                    'Principal': {
                        'Service': [
                            'codebuild.amazonaws.com',
                        ]
                    }
                }
            ]},
        )

        self.pipeline_role = Role(
            prefix + 'PipelineRole',
            Path='/',
            Policies=[Policy(
                PolicyName=prefix + 'PipelinePolicy',
                PolicyDocument={
                    'Version': '2012-10-17',
                    'Statement': [{
                        'Effect': 'Allow',
                        'Action': 'codepipeline:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'codecommit:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 's3:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'codebuild:*',
                        'Resource': '*'
                    }]
                })],
            AssumeRolePolicyDocument={'Version': '2012-10-17', 'Statement': [
                {
                    'Action': ['sts:AssumeRole'],
                    'Effect': 'Allow',
                    'Principal': {
                        'Service': [
                            'codepipeline.amazonaws.com',
                        ]
                    }
                }
            ]},
        )

        # Create a bucket to which a ci/cd pipeline will save a build output (Lambda deployment package).
        self.lambda_deployment_bucket = Bucket(
            prefix + 'LambdaDeploymentBucket',
            BucketName=prefix.lower() + '.lambda.deployment.bucket',
            AccessControl='Private',
            PublicAccessBlockConfiguration=PublicAccessBlockConfiguration(
                BlockPublicAcls=True,
                BlockPublicPolicy=True,
                IgnorePublicAcls=True,
                RestrictPublicBuckets=True
            )
        )

        # Create deployment lambda which executes deployment actions against our original AWS Lambda function.
        self.lambda_deployment_bucket_trigger = DeploymentLambda(prefix, lambda_under_deployment)

        # We need to create an explicit permission for a S3 bucket so the bucket would be able to
        # invoke deployment lambda function.
        self.invoke_permission = Permission(
            prefix + 'LambdaDeploymentPermission',
            Action='lambda:InvokeFunction',
            FunctionName=GetAtt(self.lambda_deployment_bucket_trigger.function, 'Arn'),
            Principal='s3.amazonaws.com',
            SourceArn=GetAtt(self.lambda_deployment_bucket, 'Arn')
        )

        # Add a notification configuration to our S3 bucket which makes the bucket to fire an event and invoke
        # the deployment function every time an object is created in the bucket.
        self.s3_notification_configuration = CustomS3Notification(
            prefix + 'S3NotificationConfiguration',
            ServiceToken=S3NotificationService().service_token(),
            BucketName=self.lambda_deployment_bucket.BucketName,
            NotificationConfiguration={
                'LambdaFunctionConfigurations': [
                    {
                        'LambdaFunctionArn': GetAtt(self.lambda_deployment_bucket_trigger.function, 'Arn'),
                        'Events': [
                            's3:ObjectCreated:*',
                        ],
                    },
                ]
            },
            DependsOn=[
                self.lambda_deployment_bucket.title,
                self.invoke_permission.title,
                self.lambda_deployment_bucket_trigger.function.title
            ]
        )

        # Create a main git repository which fires a ci/cd pipeline every time a code is committed to it.
        self.git_repository = Repository(
            prefix + 'GitRepository',
            RepositoryName=prefix.lower()
        )

        self.code_build_project = Project(
            prefix + 'CodeBuildProject',
            Name=prefix + 'CodeBuildProject',
            Artifacts=Artifacts(
                Type='CODEPIPELINE'
            ),
            Environment=Environment(
                ComputeType='BUILD_GENERAL1_SMALL',
                Image='aws/codebuild/standard:2.0',
                Type='LINUX_CONTAINER',
                EnvironmentVariables=[],
            ),
            ServiceRole=GetAtt(self.code_build_role, 'Arn'),
            Source=Source(
                Type='CODEPIPELINE',
                BuildSpec=self.__read_buildspec()
            )
        )

        # Our main pipeline for ci/cd.
        self.pipeline = Pipeline(
            prefix + 'Pipeline',
            ArtifactStore=ArtifactStore(
                Location=artifacts_bucket.BucketName,
                Type='S3'
            ),
            Name=prefix + 'Pipeline',
            RoleArn=GetAtt(self.pipeline_role, 'Arn'),
            Stages=[
                Stages(
                    Name='SourceStage',
                    Actions=[
                        Actions(
                            Name='SourceAction',
                            ActionTypeId=ActionTypeId(
                                Category='Source',
                                Owner='AWS',
                                Version='1',
                                Provider='CodeCommit'
                            ),
                            OutputArtifacts=[
                                OutputArtifacts(
                                    Name='SourceOutput'
                                )
                            ],
                            Configuration={
                                'RepositoryName': self.git_repository.RepositoryName,
                                'BranchName': 'master'
                            },
                            RunOrder='1'
                        )
                    ]
                ),
                Stages(
                    Name='BuildStage',
                    Actions=[
                        Actions(
                            Name='BuildAction',
                            ActionTypeId=ActionTypeId(
                                Category='Build',
                                Owner='AWS',
                                Version='1',
                                Provider='CodeBuild'
                            ),
                            InputArtifacts=[
                                InputArtifacts(
                                    Name='SourceOutput'
                                )
                            ],
                            OutputArtifacts=[
                                OutputArtifacts(
                                    Name='BuildOutput'
                                )
                            ],
                            Configuration={
                                'ProjectName': self.code_build_project.Name,
                            },
                            RunOrder='1'
                        )
                    ]
                ),
                Stages(
                    Name='DeployStage',
                    Actions=[
                        Actions(
                            Name='DeployAction',
                            ActionTypeId=ActionTypeId(
                                Category='Deploy',
                                Owner='AWS',
                                Version='1',
                                Provider='S3'
                            ),
                            InputArtifacts=[
                                InputArtifacts(
                                    Name='BuildOutput'
                                )
                            ],
                            Configuration={
                                'BucketName': self.lambda_deployment_bucket.BucketName,
                                'Extract': False,
                                'ObjectKey': 'package.zip'
                            },
                            RunOrder='1'
                        )
                    ]
                )
            ],
            DependsOn=[
                artifacts_bucket.title,
                self.git_repository.title,
                self.code_build_project.title,
                lambda_under_deployment.title
            ]
        )

    def add(self, template: Template) -> None:
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        self.lambda_deployment_bucket_trigger.add(template)

        template.add_resource(self.pipeline_role)
        template.add_resource(self.code_build_role)
        template.add_resource(self.deployment_group_role)
        template.add_resource(self.lambda_deployment_bucket)
        template.add_resource(self.invoke_permission)
        template.add_resource(self.s3_notification_configuration)
        template.add_resource(self.git_repository)
        template.add_resource(self.code_build_project)
        template.add_resource(self.pipeline)

    def __read_buildspec(self) -> str:
        """
        Reads and returns buildspec config for build project..

        :return: Buildspec config as a string.
        """
        runtime = self.RUNTIME_TO_BUILD_SPECS_MAP.get(self.lambda_under_deployment.Runtime)
        if not runtime:
            raise ValueError(f'This runtime ({self.lambda_under_deployment.Runtime}) is currently not supported.')

        with open(f'{lambda_ci_cd_root}/{runtime}', 'r') as file:
            return ''.join(file.readlines())



from typing import List, Any, Dict
from troposphere import GetAtt, Template, Ref
from troposphere.awslambda import Code, Function, VPCConfig, Environment
from troposphere.ec2 import SecurityGroup, Subnet
from troposphere.iam import Role


class LambdaFunction:
    """
    Class which creates our main AWS Lambda function.
    """
    def __init__(
            self,
            prefix: str,
            description: str,
            memory: int,
            timeout: int,
            handler: str,
            runtime: str,
            role: Role,
            env: Dict[Any, Any],
            security_groups: List[SecurityGroup],
            subnets: List[Subnet]
    ) -> None:
        """
        Constructor.

        :param prefix: Prefix string for function name.
        :param description: Function description.
        :param memory: Memory units (e.g. 128, 256, 512...) for the function.
        :param timeout: Time in seconds after which a function will be halted.
        :param handler: Method name of the function to call.
        :param runtime: Runtime environment (e.g. python3.6, nodejs10.0).
        :param role: Role to attach to the function.
        :param env: OS-level environment variables for the function.
        :param security_groups: Security groups for the function.
        :param subnets: Subnets in which the function lives. Note, subnets must have NAT Gateway.
        """
        self.lambda_function = Function(
            prefix + "Lambda",
            Code=Code(ZipFile=' '),
            Handler=handler,
            Role=GetAtt(role, "Arn"),
            Runtime=runtime,
            MemorySize=memory,
            FunctionName=prefix + 'Lambda',
            Timeout=timeout,
            Environment=Environment(
                Variables=env
            ),
            Description=description,
            VpcConfig=VPCConfig(
                SecurityGroupIds=[Ref(sg) for sg in security_groups],
                SubnetIds=[Ref(sub) for sub in subnets]
            )
        )

    def add(self, template: Template) -> None:
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        template.add_resource(self.lambda_function)



import json
import logging
import boto3

from typing import Any, Dict

client = boto3.client('s3')
logger = logging.getLogger()
logger.setLevel(logging.INFO)


class Notification:
    @staticmethod
    def create(**kwargs) -> Dict[str, Any]:
        kwargs = dict(
            Bucket=kwargs.get('bucketName'),
            NotificationConfiguration=kwargs.get('notificationConfiguration')
        )

        kwargs = {key: value for key, value in kwargs.items() if key and value}

        logger.info(
            f'Calling aws with given arguments: '
            f'{json.dumps(kwargs, default=lambda o: "<not serializable>")}.'
        )

        return client.put_bucket_notification_configuration(**kwargs)



from troposphere import AWSObject
from troposphere.validators import boolean


class CustomEcsService(AWSObject):
    resource_type = "Custom::EcsService"

    props = {
        'ServiceToken': (str, True),
        'Cluster': (str, False),
        'ServiceName': (str, True),
        'TaskDefinition': (str, False),
        'LoadBalancers': ([dict], False),
        'ServiceRegistries': ([dict], False),
        'DesiredCount': (int, False),
        'LaunchType': (str, False),
        'PlatformVersion': (str, False),
        'Role': (str, False),
        'DeploymentConfiguration': (dict, False),
        'PlacementConstraints': ([dict], False),
        'PlacementStrategy': ([dict], False),
        'NetworkConfiguration': (dict, False),
        'HealthCheckGracePeriodSeconds': (int, False),
        'SchedulingStrategy': (str, False),
        'DeploymentController': (dict, False),
        'Tags': ([dict], False),
        'EnableECSManagedTags': (boolean, False),
        'PropagateTags': (str, False),
    }



from typing import List
from troposphere.certificatemanager import Certificate
from troposphere.ec2 import SecurityGroup, Subnet, VPC
from troposphere.elasticloadbalancingv2 import TargetGroup, LoadBalancer, Listener, Action, RedirectConfig, Certificate as LBCertificate
from troposphere import Template, Ref, Output, GetAtt

class Loadbalancing:
    """
    Class which creates a loadbalancer for ecs.
    """
    TARGET_GROUP_PORT = 80
    LISTENER_HTTP_PORT_1 = 80
    LISTENER_HTTPS_PORT_1 = 443
    LISTENER_HTTP_PORT_2 = 8000
    LISTENER_HTTPS_PORT_2 = 44300

    def __init__(
            self,
            prefix: str,
            lb_security_groups: List[SecurityGroup],
            subnets: List[Subnet],
            vpc: VPC,
            desired_domain_name: str
    ):
        """
        Constructor.

        :param prefix: A prefix for resource names.
        :param lb_security_groups: Security groups to attach to a loadbalancer. NOTE! when passing loadbalancer
        security groups - make sure the loadbalancer can communicate through ci/cd blue/green deployments
        opened ports. Usually they are 8000 and 44300.
        :param subnets: Subnets in which loadbalancer can exist.
        :param vpc: Virtual private cloud in which target groups and a loadbalancer exist.
        :param desired_domain_name: Domain name for using https.
        """
        # If your service's task definition uses the awsvpc network mode
        # (which is required for the Fargate launch type), you must choose ip as the target type,
        # not instance, when creating your target groups because
        # tasks that use the awsvpc network mode are associated with an elastic network interface,
        # not an Amazon EC2 instance.
        self.target_type = 'ip'

        # Certificate so a loadbalancer could communicate via HTTPS.
        self.certificate = Certificate(
            prefix + 'FargateEcsCertificate',
            DomainName=desired_domain_name,
            ValidationMethod='DNS',
        )

        # A main target group to which a loadbalancer forwards a HTTP traffic.
        # This is the main group with which our ecs container is associated.
        self.target_group_1_http = TargetGroup(
            prefix + 'FargateEcsTargetGroup1',
            Name=prefix + 'FargateEcsTargetGroup1',
            Port=self.TARGET_GROUP_PORT,
            Protocol='HTTP',
            VpcId=Ref(vpc),
            TargetType=self.target_type
        )

        # Second target group is usd for Blue/Green deployments. A new container (that should be deployed)
        # is associated with the second target group.
        self.target_group_2_http = TargetGroup(
            prefix + 'FargateEcsTargetGroup2',
            Name=prefix + 'FargateEcsTargetGroup2',
            Port=self.TARGET_GROUP_PORT,
            Protocol='HTTP',
            VpcId=Ref(vpc),
            TargetType=self.target_type
        )

        self.load_balancer = LoadBalancer(
            prefix + 'FargateEcsLoadBalancer',
            Subnets=[Ref(sub) for sub in subnets],
            SecurityGroups=[Ref(group) for group in lb_security_groups],
            Name=prefix + 'FargateEcsLoadBalancer',
            Scheme='internet-facing',
        )

        self.load_balancer_output = Output(
            prefix + 'FargateEcsLoadBalancerUrl',
            Description='The endpoint url of a loadbalancer.',
            Value=GetAtt(self.load_balancer, 'DNSName')
        )

        # Listener that listens to HTTP incoming traffic and redirects to other HTTPS listener.
        self.listener_http_1 = Listener(
            prefix + 'FargateEcsHttpListener1',
            Port=self.LISTENER_HTTP_PORT_1,
            Protocol='HTTP',
            LoadBalancerArn=Ref(self.load_balancer),
            DefaultActions=[
                # Redirect to https.
                Action(
                    Type='redirect',
                    RedirectConfig=RedirectConfig(
                        Host='#{host}',
                        Path='/#{path}',
                        Port=str(self.LISTENER_HTTPS_PORT_1),
                        Query='#{query}',
                        StatusCode='HTTP_301',
                        Protocol='HTTPS'
                    )
                )
            ]
        )

        # Listener that listens to HTTPS traffic and forwards to a target group.
        self.listener_https_1 = Listener(
            prefix + 'FargateEcsHttpsListener1',
            Certificates=[LBCertificate(
                CertificateArn=Ref(self.certificate)
            )],
            Port=self.LISTENER_HTTPS_PORT_1,
            Protocol='HTTPS',
            LoadBalancerArn=Ref(self.load_balancer),
            DefaultActions=[
                Action(
                    Type='forward',
                    TargetGroupArn=Ref(self.target_group_1_http)
                )
            ]
        )

        # Second listener is usd for Blue/Green deployments (testing new instance). Test HTTP traffic is
        # redirected to test HTTPS traffic.
        self.listener_http_2 = Listener(
            prefix + 'FargateEcsHttpListener2',
            Port=self.LISTENER_HTTP_PORT_2,
            Protocol='HTTP',
            LoadBalancerArn=Ref(self.load_balancer),
            DefaultActions=[
                # Redirect to https.
                Action(
                    Type='redirect',
                    RedirectConfig=RedirectConfig(
                        Host='#{host}',
                        Path='/#{path}',
                        Port=str(self.LISTENER_HTTPS_PORT_2),
                        Query='#{query}',
                        StatusCode='HTTP_301',
                        Protocol='HTTPS'
                    )
                )
            ]
        )

        # Listener that listens to test HTTP traffic and forwards to a secondary target group (new container).
        self.listener_https_2 = Listener(
            prefix + 'FargateEcsHttpsListener2',
            Certificates=[LBCertificate(
                CertificateArn=Ref(self.certificate)
            )],
            Port=self.LISTENER_HTTPS_PORT_2,
            Protocol='HTTPS',
            LoadBalancerArn=Ref(self.load_balancer),
            DefaultActions=[
                Action(
                    Type='forward',
                    TargetGroupArn=Ref(self.target_group_2_http)
                )
            ]
        )

    def add(self, template: Template):
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        template.add_resource(self.certificate)

        template.add_resource(self.target_group_1_http)
        template.add_resource(self.target_group_2_http)

        template.add_resource(self.listener_http_1)
        template.add_resource(self.listener_https_1)
        template.add_resource(self.listener_http_2)
        template.add_resource(self.listener_https_2)

        template.add_resource(self.load_balancer)

        template.add_output(self.load_balancer_output)



from troposphere.applicationautoscaling import ScalableTarget, ScalingPolicy, TargetTrackingScalingPolicyConfiguration
from troposphere.applicationautoscaling import PredefinedMetricSpecification
from troposphere.iam import Role, Policy
from troposphere import Template, Ref, GetAtt, Join


class Autoscaling:
    """
    Class that creates an ECS autoscaling.
    """
    def __init__(self, prefix: str, cluster_name: str, service_name: str, service_resource_name: str):
        """
        Container.

        :param prefix: Prefix for newly created resources.
        :param cluster_name: The name of the ECS cluster.
        :param service_name: The name of the ECS service.
        :param service_resource_name: The name of the ECS service resource to specify the dependency. Autoscaling
        can not be created until an ECS service is created.
        """
        self.autoscaling_role = Role(
            prefix + 'FargateEcsAutoscalingRole',
            Path='/',
            Policies=[Policy(
                PolicyName=prefix + 'FargateEcsAutoscalingRole',
                PolicyDocument={
                    'Version': '2012-10-17',
                    'Statement': [{
                        "Action": ["logs:*"],
                        "Resource": "arn:aws:logs:*:*:*",
                        "Effect": "Allow"
                    }]
                })],
            AssumeRolePolicyDocument={'Version': '2012-10-17', 'Statement': [
                {
                    'Action': ['sts:AssumeRole'],
                    'Effect': 'Allow',
                    'Principal': {
                        'Service': [
                            'ecs.amazonaws.com',
                        ]
                    }
                }
            ]},
        )

        self.scalable_target = ScalableTarget(
            prefix + 'FargateEcsScalableTarget',
            MinCapacity=1,
            MaxCapacity=5,
            ServiceNamespace='ecs',
            DependsOn=service_resource_name,

            # The Amazon Resource Name (ARN) of an AWS Identity and Access Management (IAM) role that
            # allows Application Auto Scaling to modify the scalable target on your behalf.
            RoleARN=GetAtt(self.autoscaling_role, 'Arn'),

            # The identifier of the resource that is associated with the scalable target.
            # This string consists of the resource type and unique identifier.
            # ECS service - The resource type is service and the unique identifier
            # is the cluster name and service name. Example: service/default/sample-webapp.
            ResourceId=Join(delimiter='/', values=[
                'service',
                cluster_name,
                service_name
            ]),
            # Read more on scalable dimensions:
            # https://docs.aws.amazon.com/autoscaling/application/APIReference
            # /API_RegisterScalableTarget.html#autoscaling-RegisterScalableTarget-request-ScalableDimension
            ScalableDimension='ecs:service:DesiredCount'
        )

        self.scaling_policy = ScalingPolicy(
            prefix + 'FargateEcsScalingPolicy',
            PolicyName=prefix + 'FargateEcsScalingPolicy',
            ScalingTargetId=Ref(self.scalable_target),
            PolicyType='TargetTrackingScaling',
            TargetTrackingScalingPolicyConfiguration=TargetTrackingScalingPolicyConfiguration(
                DisableScaleIn=False,
                TargetValue=75.0,
                PredefinedMetricSpecification=PredefinedMetricSpecification(
                    PredefinedMetricType='ECSServiceAverageCPUUtilization'
                )
            )
        )

    def add(self, template: Template):
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        template.add_resource(self.autoscaling_role)
        template.add_resource(self.scalable_target)
        template.add_resource(self.scaling_policy)



from typing import Union
from troposphere.ecs import Service, Cluster
from troposphere.s3 import Bucket
from aws_infrastructure_sdk.cloud_formation.custom_resources.resource.ecs_service import CustomEcsService
from aws_infrastructure_sdk.cloud_formation.custom_resources.resource.git_commit import CustomGitCommit
from aws_infrastructure_sdk.cloud_formation.custom_resources.service.deployment_group import DeploymentGroupService
from aws_infrastructure_sdk.cloud_formation.custom_resources.service.git_commit import GitCommitService
from aws_infrastructure_sdk.cloud_formation.types import AwsRef
from troposphere.codecommit import Repository as GitRepository
from troposphere.ecr import Repository as EcrRepository
from troposphere.codedeploy import Application
from troposphere.elasticloadbalancingv2 import TargetGroup, Listener
from troposphere.iam import Role, Policy
from troposphere import Template, GetAtt, Ref, Output, Join
from troposphere.codepipeline import *
from aws_infrastructure_sdk.cloud_formation.custom_resources.resource.deployment_group import CustomDeploymentGroup


class EcsPipeline:
    """
    Class which creates infrastructure for CI/CD Blue/Green Ecs Fargate deployment.
    """
    def __init__(
            self,
            prefix: str,
            aws_account_id: str,
            aws_region: str,
            main_target_group: TargetGroup,
            deployments_target_group: TargetGroup,
            main_listener: Listener,
            deployments_listener: Listener,
            ecs_service: Union[CustomEcsService, Service],
            ecs_cluster: Cluster,
            artifact_builds_s3: Bucket,
            task_def: AwsRef,
            app_spec: AwsRef,
    ):
        """
        Constructor.

        :param prefix: A prefix for newly created resources.
        :param aws_account_id: An account id under which a CF stack is running.
        :param aws_region: Region in which the CF stack is running.
        :param main_target_group: A target group to which a loadbalancer is forwarding a received production traffic.
        :param deployments_target_group: A target group to which a loadbalancer is forwarding a received test traffic.
        :param main_listener: A listener which receives incoming traffic and forwards it to a target group.
        :param deployments_listener: A listener which receives incoming traffic and forwards it to a target group.
        This listener is used for blue/green deployment.
        :param ecs_service: Ecs service to which create this pipeline.
        :param ecs_cluster: ECS cluster in which the ECS service is.
        :param artifact_builds_s3: A S3 bucket to which built artifacts are written.
        :param task_def: Task definition object defining the parameters for a newly deployed container.
        :param app_spec: App specification object defining the ecs service modifications.
        """
        self.deployment_group_role = Role(
            prefix + 'FargateEcsDeploymentGroupRole',
            Path='/',
            Policies=[Policy(
                PolicyName=prefix + 'FargateEcsDeploymentGroupPolicy',
                PolicyDocument={
                    'Version': '2012-10-17',
                    'Statement': [{
                        'Action': [
                            "ecs:DescribeServices",
                            "ecs:CreateTaskSet",
                            "ecs:UpdateServicePrimaryTaskSet",
                            "ecs:DeleteTaskSet",
                            "elasticloadbalancing:DescribeTargetGroups",
                            "elasticloadbalancing:DescribeListeners",
                            "elasticloadbalancing:ModifyListener",
                            "elasticloadbalancing:DescribeRules",
                            "elasticloadbalancing:ModifyRule",
                            "lambda:InvokeFunction",
                            "cloudwatch:DescribeAlarms",
                            "sns:Publish",
                            "s3:GetObject",
                            "s3:GetObjectMetadata",
                            "s3:GetObjectVersion",
                            "iam:PassRole"
                        ],
                        "Resource": "*",
                        "Effect": "Allow"
                    }]
                })],
            AssumeRolePolicyDocument={'Version': '2012-10-17', 'Statement': [
                {
                    'Action': ['sts:AssumeRole'],
                    'Effect': 'Allow',
                    'Principal': {
                        'Service': [
                            'ecs-tasks.amazonaws.com',
                            'codedeploy.amazonaws.com',
                        ]
                    }
                }
            ]},
        )

        # TODO needs permissions fixing.
        self.pipeline_role = Role(
            prefix + 'FargateEcsPipelineRole',
            Path='/',
            Policies=[Policy(
                PolicyName=prefix + 'FargateEcsPipelinePolicy',
                PolicyDocument={
                    'Version': '2012-10-17',
                    'Statement': [{
                        'Effect': 'Allow',
                        'Action': 'codepipeline:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'codecommit:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 's3:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'codebuild:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'codedeploy:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'ecs:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'ecr:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'ec2:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'iam:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'elasticloadbalancing:*',
                        'Resource': '*'
                    }, {
                        'Effect': 'Allow',
                        'Action': 'logs:*',
                        'Resource': '*'
                    }]
                })],
            AssumeRolePolicyDocument={'Version': '2012-10-17', 'Statement': [
                {
                    'Action': ['sts:AssumeRole'],
                    'Effect': 'Allow',
                    'Principal': {
                        'Service': [
                            'codepipeline.amazonaws.com',
                            'codecommit.amazonaws.com',
                            'codebuild.amazonaws.com',
                            'codedeploy.amazonaws.com',
                            'ecs-tasks.amazonaws.com',
                            'iam.amazonaws.com',
                            'ecs.amazonaws.com',
                            's3.amazonaws.com',
                            'ec2.amazonaws.com'
                        ]
                    }
                }
            ]},
        )

        self.git_repository = GitRepository(
            prefix + 'FargateEcsGitRepository',
            RepositoryDescription=(
                'Repository containing appspec and taskdef files for ecs code-deploy blue/green deployments.'
            ),
            RepositoryName=prefix.lower() + '_config'
        )

        # Commit configuration files to a git repository from which a code-pipeline will read later.
        self.commit = CustomGitCommit(
            prefix + 'FargateEcsDeploymentConfig',
            ServiceToken=GitCommitService().service_token(),
            RepositoryName=self.git_repository.RepositoryName,
            BranchName='master',
            CommitMessage='Initial appspec and taskdef files.',
            PutFiles=[
                {
                    'filePath': 'taskdef.json',
                    'fileMode': 'NORMAL',
                    'fileContent': task_def,
                }, {
                    'filePath': 'appspec.yaml',
                    'fileMode': 'NORMAL',
                    'fileContent': app_spec,
                }
            ],
            DependsOn=[self.git_repository.title]
        )

        self.ecr_repository = EcrRepository(
            prefix + 'FargateEcsEcrRepository',
            RepositoryName=prefix.lower()
        )

        self.ecr_repository_output = Output(
            prefix + 'FargateEcsEcrOutput',
            Description='Ecr endpoint to tag/push/pull docker images.',
            Value=Join(delimiter='', values=[
                f'{aws_account_id}.dkr.ecr.{aws_region}.amazonaws.com/',
                Ref(self.ecr_repository),
                f':latest',
            ])
        )

        self.application = Application(
            prefix + 'FargateEcsCodeDeployApplication',
            ApplicationName=prefix + 'FargateEcsCodeDeployApplication',
            ComputePlatform='ECS'
        )

        self.deployment_group = CustomDeploymentGroup(
            prefix + 'FargateEcsDeploymentGroup',
            ServiceToken=DeploymentGroupService().service_token(),
            ApplicationName=self.application.ApplicationName,
            DeploymentGroupName=prefix + 'FargateEcsDeploymentGroup',
            DeploymentConfigName='CodeDeployDefault.ECSAllAtOnce',
            ServiceRoleArn=GetAtt(self.deployment_group_role, 'Arn'),
            AutoRollbackConfiguration={
                'enabled': True,
                'events': ['DEPLOYMENT_FAILURE', 'DEPLOYMENT_STOP_ON_ALARM', 'DEPLOYMENT_STOP_ON_REQUEST']
            },
            DeploymentStyle={
                'deploymentType': 'BLUE_GREEN',
                'deploymentOption': 'WITH_TRAFFIC_CONTROL'
            },
            BlueGreenDeploymentConfiguration={
                'terminateBlueInstancesOnDeploymentSuccess': {
                    'action': 'TERMINATE',
                    'terminationWaitTimeInMinutes': 5
                },
                'deploymentReadyOption': {
                    'actionOnTimeout': 'CONTINUE_DEPLOYMENT',
                },
            },
            LoadBalancerInfo={
                'targetGroupPairInfoList': [
                    {
                        'targetGroups': [
                            {
                                'name': main_target_group.Name
                            },
                            {
                                'name': deployments_target_group.Name
                            },
                        ],
                        'prodTrafficRoute': {
                            'listenerArns': [
                                Ref(main_listener)
                            ]
                        },
                        'testTrafficRoute': {
                            'listenerArns': [
                                Ref(deployments_listener)
                            ]
                        }
                    },
                ]
            },
            EcsServices=[
                {
                    'serviceName': ecs_service.ServiceName,
                    'clusterName': ecs_cluster.ClusterName
                },
            ],
            DependsOn=[
                ecs_service.title,
            ]
        )

        # Create a pipeline which deploys from ECR to ECS.
        self.pipeline = Pipeline(
            prefix + 'FargateEcsPipeline',
            ArtifactStore=ArtifactStore(
                Location=artifact_builds_s3.BucketName,
                Type='S3'
            ),
            Name=prefix + 'FargateEcsPipeline',
            RoleArn=GetAtt(self.pipeline_role, 'Arn'),
            Stages=[
                Stages(
                    Name='SourceStage',
                    Actions=[
                        # Source the image from ECR.
                        Actions(
                            Name='SourceEcrAction',
                            ActionTypeId=ActionTypeId(
                                Category='Source',
                                Owner='AWS',
                                Version='1',
                                Provider='ECR'
                            ),
                            OutputArtifacts=[
                                OutputArtifacts(
                                    Name='EcsImage'
                                )
                            ],
                            Configuration={
                                'RepositoryName': self.ecr_repository.RepositoryName
                            },
                            RunOrder='1'
                        ),
                        # Source configuration from git repository.
                        Actions(
                            Name='SourceCodeCommitAction',
                            ActionTypeId=ActionTypeId(
                                Category='Source',
                                Owner='AWS',
                                Version='1',
                                Provider='CodeCommit'
                            ),
                            OutputArtifacts=[
                                OutputArtifacts(
                                    Name='EcsConfig'
                                )
                            ],
                            Configuration={
                                'RepositoryName': self.git_repository.RepositoryName,
                                'BranchName': 'master'
                            },
                            RunOrder='1'
                        )
                    ]
                ),
                Stages(
                    Name='DeployStage',
                    Actions=[
                        # Initiate the deployment.
                        Actions(
                            Name='DeployAction',
                            ActionTypeId=ActionTypeId(
                                Category='Deploy',
                                Owner='AWS',
                                Version='1',
                                Provider='CodeDeployToECS'
                            ),
                            InputArtifacts=[
                                InputArtifacts(
                                    Name='EcsImage'
                                ),
                                InputArtifacts(
                                    Name='EcsConfig'
                                )
                            ],
                            Configuration={
                                "TaskDefinitionTemplateArtifact": "EcsConfig",
                                "AppSpecTemplateArtifact": "EcsConfig",

                                "TaskDefinitionTemplatePath": "taskdef.json",
                                "AppSpecTemplatePath": "appspec.yaml",

                                "ApplicationName": self.application.ApplicationName,
                                "DeploymentGroupName": self.deployment_group.DeploymentGroupName,

                                "Image1ArtifactName": "EcsImage",
                                "Image1ContainerName": "IMAGE1_NAME"
                            },
                            RunOrder='1'
                        )
                    ]
                )
            ],
            # The pipeline can not be created until applications, deployment groups, git repositories and ecs
            # services are created.
            DependsOn=[
                artifact_builds_s3.title,
                self.application.title,
                self.deployment_group.title,
                self.git_repository.title,
                self.ecr_repository.title,
            ]
        )

    def add(self, template: Template):
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        template.add_resource(self.commit)
        template.add_resource(self.git_repository)
        template.add_resource(self.ecr_repository)
        template.add_resource(self.deployment_group_role)
        template.add_resource(self.pipeline_role)
        template.add_resource(self.application)
        template.add_resource(self.deployment_group)
        template.add_resource(self.pipeline)

        template.add_output(self.ecr_repository_output)



from multiprocessing import Process
from typing import List
from aws_infrastructure_sdk.cloud_formation.custom_resources.config import CF_CUSTOM_BUILDS_BUCKET
from aws_infrastructure_sdk.cloud_formation.custom_resources.service.abstract_custom_service import AbstractCustomService
from aws_infrastructure_sdk.lambdas.deployment.deployment_package import DeploymentPackage


class CustomResourcesBuilder:
    def __init__(self, region: str, profile: str, suffix: str):
        self.suffix = suffix
        self.region = region
        self.profile = profile

    def build(self, custom_resource_services: List[AbstractCustomService], upload: bool = True):
        process_pool = []

        for resource in custom_resource_services:
            p = Process(target=self.__build_single, args=(resource, upload))
            p.start()

            process_pool.append(p)

        for process in process_pool:
            process.join(600 if upload else 500)

    def __build_single(self, resource: AbstractCustomService, upload: bool = True):
        DeploymentPackage(
            environment='none',
            project_src_path=resource.src,
            lambda_name=resource.lambda_name,
            s3_upload_bucket=CF_CUSTOM_BUILDS_BUCKET.format(self.suffix),
            s3_bucket_region=self.region,
            aws_profile=self.profile,
            refresh_lambda=upload
        ).deploy()



from typing import List, Dict
from troposphere.logs import LogGroup
from aws_infrastructure_sdk.cloud_formation.custom_resources.service.ecs_service import EcsServiceService
from aws_infrastructure_sdk.cloud_formation.types import AwsRef
from troposphere.ec2 import SecurityGroup, Subnet
from troposphere.elasticloadbalancingv2 import TargetGroup, Listener
from troposphere import Template, Ref, GetAtt, Join
from troposphere.ecs import *
from troposphere.iam import Role, Policy
from aws_infrastructure_sdk.cloud_formation.custom_resources.resource.ecs_service import CustomEcsService


class Ecs:
    """
    Class that creates ECS Fargate service.
    """
    def __init__(
            self,
            prefix: str,
            aws_region: str,
            cpu: str,
            ram: str,
            environment: Dict[str, AwsRef],
            container_name: str,
            container_port: int,
            target_group: TargetGroup,
            security_groups: List[SecurityGroup],
            subnets: List[Subnet],
            depends_on_loadbalancers: List[LoadBalancer] = [],
            depends_on_target_groups: List[TargetGroup] = [],
            depends_on_listeners: List[Listener] = []
    ) -> None:
        """
        Constructor.

        :param prefix: A prefix for newly created resources.
        :param aws_region: A region in which resources are put.
        :param cpu: Cpu points for the deployed container. 1 CPU = 1024 Cpu points.
        :param ram: Memory for the deployed container. 1 GB Ram = 1024.
        :param environment: Environment that will be passed to a running container.
        :param container_name: The name that will be given to a newly deployed container.
        :param container_port: An open container port through which a loadbalancer can communicate.
        :param target_group: A main target group to which a loadbalancer will forward traffic. Also, the newly
        created container will be associated with this group.
        :param security_groups: Container security groups restricting network traffic.
        :param subnets: Subnets in which the newly created container can be placed.
        :param depends_on_loadbalancers: Before creating ecs service, these loadbalancers must be created.
        :param depends_on_target_groups: Before creating ecs service, these target groups must be created.
        :param depends_on_listeners: Before creating ecs service, these listeners must be created.
        """
        self.prefix = prefix
        self.aws_region = aws_region
        self.environment = environment
        self.cpu = cpu
        self.ram = ram
        self.container_name = container_name
        self.container_port = container_port

        self.task_execution_role = Role(
            prefix + 'FargateEcsTaskExecutionRole',
            Path='/',
            Policies=[Policy(
                PolicyName=prefix + 'FargateEcsTaskExecutionPolicy',
                PolicyDocument={
                    'Version': '2012-10-17',
                    'Statement': [{
                        'Action': [
                            "ecr:GetAuthorizationToken",
                            "ecr:BatchCheckLayerAvailability",
                            "ecr:GetDownloadUrlForLayer",
                            "ecr:BatchGetImage",
                            "logs:CreateLogStream",
                            "logs:PutLogEvents"
                        ],
                        "Resource": "*",
                        "Effect": "Allow"
                    }]
                })],
            AssumeRolePolicyDocument={'Version': '2012-10-17', 'Statement': [
                {
                    'Action': ['sts:AssumeRole'],
                    'Effect': 'Allow',
                    'Principal': {
                        'Service': [
                            'ecs-tasks.amazonaws.com',
                        ]
                    }
                }
            ]},
        )

        self.log_group = LogGroup(
            prefix + 'FargateEcsLogGroup',
            LogGroupName=f'/aws/ecs/fargate/{prefix}'
        )

        self.cluster = Cluster(
            prefix + 'FargateEcsCluster',
            ClusterName=prefix + 'FargateEcsCluster'
        )

        self.task = TaskDefinition(
            prefix + 'FargateEcsTaskDefinition',
            RequiresCompatibilities=['FARGATE'],
            ExecutionRoleArn=GetAtt(self.task_execution_role, 'Arn'),
            ContainerDefinitions=[
                ContainerDefinition(
                    Name=container_name,
                    # Create dummy image, since container definitions list can not be empty.
                    Image='nginx:latest',
                    # For task definitions that use the awsvpc network mode, you should only specify the containerPort.
                    # The hostPort can be left blank or it must be the same value as the containerPort.
                    PortMappings=[
                        PortMapping(
                            ContainerPort=80
                        )
                    ],
                    LogConfiguration=LogConfiguration(
                        LogDriver='awslogs',
                        Options={
                            # Use Ref to set a dependency to a log group.
                            # Or use "depends on" attribute.
                            'awslogs-group': Ref(self.log_group),
                            'awslogs-region': aws_region,
                            'awslogs-stream-prefix': prefix
                        }
                    )
                )
            ],
            Cpu=cpu,
            Memory=ram,
            # For ECS Fargate - awsvpc is the only available option.
            NetworkMode='awsvpc',
            Family=prefix.lower()
        )

        self.service = CustomEcsService(
            prefix + 'FargateEcsService',
            ServiceToken=EcsServiceService().service_token(),
            Cluster=Ref(self.cluster),
            ServiceName=prefix + 'FargateEcsService',
            TaskDefinition=Ref(self.task),
            LoadBalancers=[
                {
                    'targetGroupArn': Ref(target_group),
                    'containerName': container_name,
                    'containerPort': container_port
                },
            ],
            DesiredCount=1,
            LaunchType='FARGATE',
            NetworkConfiguration={
                'awsvpcConfiguration': {
                    'subnets': [Ref(sub) for sub in subnets],
                    'securityGroups': [Ref(sub) for sub in security_groups],
                    'assignPublicIp': 'ENABLED'
                }
            },
            DeploymentController={
                'type': 'CODE_DEPLOY'
            },
            # Target groups must have an associated load balancer before creating an ecs service.
            DependsOn=(
                    [lb.title for lb in depends_on_loadbalancers] +
                    [tg.title for tg in depends_on_target_groups] +
                    [l.title for l in depends_on_listeners]
            )
        )

    def create_task_def(self) -> AwsRef:
        """
        Creates a task definition object which will be used for deploying new containers through a pipeline.
        The task definition object specifies parameters about newly created containers.

        :return: Task definition object.
        """
        environment_list = []

        for key, value in self.environment.items():
            join = Join(delimiter='', values=['{"name": "', key, '", "value": "', value, '"}'])
            environment_list.append(join)

        environment = Join(delimiter='\n', values=[
            '"environment": [',
            Join(delimiter=',\n', values=environment_list),
            '],'
        ])

        definition = Join(delimiter='\n', values=[
            '{',
            Join(delimiter='', values=['    "executionRoleArn": ', '"', GetAtt(self.task_execution_role, "Arn"), '"', ',']),
            '    "containerDefinitions": [',
            '        {',
            '            "name": "{}",'.format(self.container_name),
            '            "image": "<IMAGE1_NAME>",',
            '            "essential": true,',
            environment,
            # For task definitions that use the awsvpc network mode, you should only specify the containerPort.
            # The hostPort can be left blank or it must be the same value as the containerPort.
            '            "portMappings": [',
            '                {',
            '                   "containerPort": {}'.format(self.container_port),
            '                }',
            '            ],',
            '            "logConfiguration": {',
            '                "logDriver": "awslogs",',
            '                "options": {',
            '                    "awslogs-group": "{}",'.format(self.log_group.LogGroupName),
            '                    "awslogs-region": "{}",'.format(self.aws_region),
            '                    "awslogs-stream-prefix": "{}"'.format(self.prefix),
            '                }',
            '            }',
            '        }',
            '    ],',
            '    "requiresCompatibilities": [',
            '        "FARGATE"',
            '    ],',
            '    "networkMode": "awsvpc",',
            '    "cpu": "{}",'.format(self.cpu),
            '    "memory": "{}",'.format(self.ram),
            '    "family": "{}"'.format(self.prefix.lower()),
            '}'
        ])

        return definition

    def create_appspec(self) -> AwsRef:
        """
        Creates an application specification object which will be used for deploying new containers through a pipeline.
        The application specification object specifies parameters about the ECS service.

        :return: Application specification object.
        """
        app_spec = (
            f'version: 0.0',
            f'Resources:',
            f'  - TargetService:',
            f'      Type: AWS::ECS::Service',
            f'      Properties:',
            f'        TaskDefinition: <TASK_DEFINITION>',
            f'        LoadBalancerInfo:',
            f'          ContainerName: "{self.container_name}"',
            f'          ContainerPort: 80',
        )

        return '\n'.join(app_spec)

    def add(self, template: Template):
        """
        Adds all created resources to a template.

        :param template: Template to which resources should be added.

        :return: No return.
        """
        template.add_resource(self.log_group)
        template.add_resource(self.cluster)
        template.add_resource(self.task)
        template.add_resource(self.service)
        template.add_resource(self.task_execution_role)



import botocore
import cfnresponse
import boto3
import json
import logging

logger = logging.getLogger()
logger.setLevel(logging.INFO)

logger.info(f'Version of boto3 lib: {boto3.__version__}.')
logger.info(f'Version of botocore lib: {botocore.__version__}.')


def __success(event, context, data):
    logger.info('SUCCESS: {}'.format(data))
    cfnresponse.send(event, context, cfnresponse.SUCCESS, data)


def __failed(event, context, data):
    logger.info('FAIL: {}'.format(data))
    cfnresponse.send(event, context, cfnresponse.FAILED, data)


def __create(user_pool_id: str, domain_name: str):
    client = boto3.client('cognito-idp')

    response = client.create_user_pool_domain(
        Domain=domain_name,
        UserPoolId=user_pool_id
    )

    data = json.dumps(response, default=lambda o: '<not serializable>')
    logger.info(data)


def __update(user_pool_id: str, domain_name: str):
    client = boto3.client('cognito-idp')

    response = client.update_user_pool_domain(
        Domain=domain_name,
        UserPoolId=user_pool_id
    )

    data = json.dumps(response, default=lambda o: '<not serializable>')
    logger.info(data)


def __delete(user_pool_id: str, domain_name: str):
    client = boto3.client('cognito-idp')

    try:
        response = client.delete_user_pool_domain(
            Domain=domain_name,
            UserPoolId=user_pool_id
        )

        data = json.dumps(response, default=lambda o: '<not serializable>')
        logger.info(data)
    except Exception as ex:
        logger.error(repr(ex))


def __handle(event, context):
    logger.info('Got new request. Event: {}, Context: {}'.format(event, context))

    user_pool_id = event['ResourceProperties']['UserPoolId']
    domain_name = event['ResourceProperties']['DomainName']

    if event['RequestType'] == 'Delete':
        __delete(user_pool_id, domain_name)
        return

    if event['RequestType'] == 'Create':
        __create(user_pool_id, domain_name)
        return

    if event['RequestType'] == 'Update':
        __update(user_pool_id, domain_name)
        return

    raise KeyError('Unsupported request type! Type: {}'.format(event['RequestType']))


def handler(event, context):
    try:
        __handle(event, context)
    except Exception as ex:
        __failed(event, context, {'Error': str(ex)})
        return

    __success(event, context, {'Success': True})
